<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.1/1829459
 */

namespace Api\Controller;

use Api\AbstractApiController;
use Application\Config\ConfigManager;
use Application\View\Helper\Avatar;
use Groups\Model\Group;
use P4\Exception;
use P4\Filter\Utf8 as Utf8Filter;
use P4\Log\Logger;
use Projects\Model\Project;
use Users\Model\User;
use Zend\View\Model\JsonModel;

/**
 * Basic API controller providing a simple version action
 *
 * @SWG\Resource(
 *   apiVersion="v9",
 *   basePath="/api/v9/"
 * )
 */
class UsersController extends AbstractApiController
{
    const AVATAR = "avatar";
    /**
     * @SWG\Api(
     *     path="users/",
     *     description="Gives you a list of Users from Perforce.",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Get List of Users",
     *         notes="Returns a list of users in Swarm.",
     *         nickname="listUsers",
     *         @SWG\Parameter(
     *             name="fields",
     *             description="An optional comma-separated list (or array) of fields to show for each user.
     *                          Omitting this parameter or passing an empty value shows all fields.
     *                          Beaware the fields are case sensitive for users. You can use one of the below:
     *                          User, Type, Email, Update, Access, FullName, JobView, Password, AuthMethod,
     *                          Reviews",
     *             paramType="query",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="users",
     *             description="An optional comma-separated list (or array) of users to display.
     *                          Omitting this parameter or passing an empty value shows all users.",
     *             paramType="query",
     *             type="string",
     *             required=false
     *         ),
     *         @SWG\Parameter(
     *             name="group",
     *             description="An optional to get users from a group. Cannot be used with users parameter",
     *             paramType="query",
     *             type="string",
     *             required=false
     *         )
     *     )
     * )
     *
     * @apiUsageExample Listing projects
     *
     *   To list all projects:
     *
     *   ```bash
     *   curl -u "username:password" "https://my-swarm-host/api/v9/users?fields=User,FullName&users=super,bruno"
     *   ```
     *
     *   Pagination is not currently supported by this endpoint. Swarm responds with a list of all users:
     *
     *   ```json
     *   {
     *       {
     *         "User": "bruno",
     *         "FullName": "Bruno First"
     *       },
     *       {
     *         "User": "super",
     *         "FullName": "Super Second"
     *       }
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *   {
     *       "User": "bruno",
     *       "Type": "standard",
     *       "Email": "bruno@perforce.com",
     *       "Updated": "2005/10/11 21:05:15",
     *       "Access": "2018/04/12 14:55:10",
     *       "FullName": "Bruno First",
     *       "JobView": null,
     *       "Password": null,
     *       "AuthMethod": "perforce",
     *       "Reviews": []
     *   }
     *
     * @throws \Application\Config\ConfigException
     * @return mixed
     */
    public function getList()
    {
        $request    = $this->getRequest();
        $services   = $this->services;
        $fields     = $request->getQuery('fields');
        $fetchUsers = $request->getQuery('users');
        $group      = $request->getQuery('group');
        $current    = $request->getQuery('current', false);
        $p4Admin    = $services->get('p4_admin');
        $config     = $services->get('config');

        // If we are requesting for just current user
        if ($current) {
            return $this->currentUser($services);
        }

        // If required_login is set to true we shouldn't allow non logged in user to access users.
        if (ConfigManager::getValue($config, ConfigManager::SECURITY_REQUIRE_LOGIN) === true) {
            $services->get('permissions')->enforce('authenticated');
        }

        // if requested, get only users for a specified group
        $groupUsers = $group ? Group::fetchAllMembers($group, null, null, null, $p4Admin) : null;

        $utf8    = new Utf8Filter;
        $users   = array();
        $options = null;
        if ($fetchUsers) {
            $options = array(User::FETCH_BY_NAME => is_string($fetchUsers) ? explode(',', $fetchUsers) : $fetchUsers);
        } elseif ($group) {
            $options = array( User::FETCH_BY_NAME => $groupUsers);
        }

        // check if the server is case sensitive
        $caseSensitive = $p4Admin->isCaseSensitive();

        foreach (User::fetchAll($options, $p4Admin) as $user) {
            // if the user id is on the user blacklist
            // and we have enabled mentions,
            // do not include that user in the possible mentions list
            if ($config['mentions']['mode'] != 'disabled' && $this->getRequest()->getQuery('respectBlacklist')) {
                $blacklisted = array_values($config['mentions']['usersBlacklist']);
                // check against all values in usersBlacklist
                foreach ($blacklisted as $pattern) {
                    $matches = preg_match("/$pattern/" . ($caseSensitive ? '' : 'i') . "", $user->getId());
                    if ($matches) {
                        // skip this user
                        continue 2;
                    }
                }
            }

            // though unexpected, some fields (User or FullName) can include invalid UTF-8 sequences
            // so we filter them, otherwise json encoding could crash with an error
            $data      = array();
            $badFields = array();
            $message   = "";
            $fields    = $fields ? is_string($fields) ? explode(',', $fields) : $fields : $user->getFields();
            foreach ($fields as $field) {
                try {
                    if ($field === self::AVATAR) {
                        $data[$field] = Avatar::getAvatarDetails($config, $user->getId(), $user->getEmail());
                    } else {
                        $data[ $field ] = $utf8->filter($user->get($field));
                    }
                } catch (Exception $e) {
                    // We have encountered an fields we don't know about.
                    $badFields[] = $field;
                    $message     = $e->getMessage();
                    unset($e);
                }
            }
            if (!empty($badFields)) {
                $stringBadFields = implode(', ', $badFields);
                $response        = $this->getResponse();
                $response->setStatusCode(400);
                return new JsonModel(
                    array(
                        'isValid' => false,
                        'message' =>  $message. " [".$stringBadFields."]",
                        'code'    => 400
                    )
                );
            }

            $users[] = $data;
        }

        return $this->getResponse()->isOk()
            ? $this->prepareSuccessModel($users, $fields)
            : $this->prepareErrorModel($users);
    }

    /**
     * Check if the user is logged in and then return the user object
     *
     * @param $services
     *
     * @return JsonModel
     */
    private function currentUser($services)
    {
        $error = $authUser = false;
        try {
            $p4User   = $services->get('p4_user');
            $authUser = $p4User->isAuthenticated()
                ? User::fetchById($p4User->getUser(), $p4User)->getValuesArray()
                : false;
        } catch (\Exception $e) {
            if (strpos($e->getMessage(), "Service with name \"p4_user\" could not be created") === 0) {
                $error = "User not logged in";
            } else {
                $error = $e->getMessage();
            }
        }

        return new JsonModel(
            array(
                'isValid'  => $authUser ? true : false,
                'messages' => $error ? array($error) : array(),
                'user'     => $authUser,
            )
        );
    }

    /**
     * @SWG\Api(
     *     path="users/{user}/unfollowall",
     *     description="Unfollow all Users and Projects that the given user is following",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="Unfollow all Users and Projects",
     *         notes="Admin and super users are permitted to execute unfollow all against any target
     *                user. Other users are only permitted to execute the call if they themselves
     *                are the target user"
     *     )
     * )
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *         "isValid": true,
     *         "messages": "User {user} is no longer following any Projects or Users."
     *     }
     *
     * @return  JsonModel
     * @throws \Exception
     */
    public function unfollowAllAction()
    {
        // only allow logged in users to unfollow
        $services = $this->services;
        $services->get('permissions')->enforce('authenticated');
        $translator = $services->get('translator');
        $isValid    = false;

        // Get the admin and current logged in user.
        $p4Admin     = $services->get('p4_admin');
        $currentUser = $services->get('user');
        $user        = $this->getEvent()->getRouteMatch()->getParam('user');

        // Check if the user that is being attempted to process exist else return.
        if (User::exists($user, $p4Admin)) {
            $user   = User::fetchById($user, $p4Admin);
            $userId = $user->getId();
        } else {
            $messages = $translator->t("User '%s' does not exist", array($user));
            return new JsonModel(
                array(
                    'isValid'  => false,
                    'messages' => $messages,
                )
            );
        }

        // Check if the current logged in user is on their own page else check if admin.
        if ($currentUser->getId() !== $userId) {
            $services->get('permissions')->enforce('admin');
        }

        try {
            // Get the followed projects
            $config   = $user->getConfig();
            $projects = Project::fetchAll(
                array( Project::FETCH_COUNT_FOLLOWERS => true ),
                $p4Admin
            );
            // filter out projects not accessible to the current user
            $projects = $services->get('projects_filter')->filter($projects);
            // Now filter the projects to only the projects we are following.
            $projects->filterByCallback(
                function (Project $project) use ($userId) {
                    return $project->isFollowing($userId) === true;
                }
            );
            // Fetch the following users as well.
            $followingUsers = $config->getFollows('user');
            // Now remove all the users that this user is following
            foreach ($followingUsers as $user) {
                $config->removeFollow($user, 'user')->save();
            }
            // Now remove all the projects this user is following.
            foreach ($projects as $project) {
                $config->removeFollow($project->getId(), 'project')->save();
            }
            // Now check that all user and projects are removed
            $projects->filterByCallback(
                function (Project $project) use ($userId) {
                    return $project->isFollowing($userId) === true;
                }
            );
            // Fetch the following users as well.
            $followingUsers = $config->getFollows('user');
            // Now check that we have removed all following items and set isValid to true.
            if (count($followingUsers) === 0 && count($projects) === 0) {
                $isValid = true;
            }
        } catch (Exception $error) {
            // Do nothing right now.
            Logger::log(Logger::ERR, "UserAPI: UnfollowAll : We ran into a problem :" . $error->getMessage());
            throw new \RuntimeException($error->getMessage());
        }
        $messages = $translator->t("User %s is no longer following any Projects or Users.", array($userId));
        return new JsonModel(
            array(
                'isValid'  => $isValid,
                'messages' => $messages,
            )
        );
    }
}
